/*--------------------------------------------------------------------------
  Filename:	    sgdiskmon.c
  Author:  	    Andy Cress   <arcress@users.sourceforge.net>
  Operating System: Linux 2.2 or greater
  Abstract: 	    SCSI Generic Disk Monitor service
   		    Scan the scsi bus for every active device, and 
		    recognize when a device has been failed, removed, or
                    hot-inserted.
		    Also poll the hot-swap backplane with SAF-TE commands 
		    to check for insertion events.
                    This program uses a script (sgevt) to perform actions for 
                    these events.

  ----------- Description --------------------------------------------------
  sgdiskmon
  What this service does:
  This service monitors the scsi bus to see when a new disk has been inserted.
  or removed.  It polls each possible device every 5 seconds to see if any
  device has been added or removed.  It will also poll the hot-swap backplane 
  via the SAF-TE interface, if there is a backplane/processor present.

  It does not handle cases of drives being inserted for which the kernel
  does not have a device node (new scsi id).  This requires an accompanying 
  scsi_rescan function to be initiated against the kernel.

  Normally only one disk at a time would be removed or inserted.  However,
  if two new disks were inserted, one disk would be inserted on one pass, 
  and the other disk would be inserted on the next pass.

  ----------- Change History -----------------------------------------------
  11/16/06  0.1  ARCress  created from sgraidmon.c
  03/12/07  1.1  ARCress  split to use sgsubmon.c for common routines

----------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------
Copyright (c) 2006-2008, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

  a.. Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer. 
  b.. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution. 
  c.. Neither the name of Intel Corporation nor the names of its contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  -------------------------------------------------------------------------*/
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <getopt.h>
#include <time.h>
#include <stdarg.h>
#include <sys/ioctl.h>
#include <scsi/sg.h>

#include "sgsubmon.h"

//////////////////////////////////////////////////////////////////////
// Global Data 
//////////////////////////////////////////////////////////////////////
char *progver  = "1.66";         /* program version */
char *progname    = "sgdiskmon";   /* program name */
char sgevtcmd[]   = "/sbin/sgevt";
char initscript[] = "/etc/init.d/sgdisk";
char logfile[]    = "/var/log/sgdiskmon.log";     /* log filename */   
FILE *fdlog = NULL;             /* log file descriptor */
FILE *fdmsg = NULL;             /* message file descriptor (stdout) */
char output[MSG_LEN];		/* output buffer for text messages */
char output2[MSG_LEN];		/* 2nd output buffer for messages */
char fdebug = 0;                /* =1 for debug messages */
char frunonce = 0;              /* =1 to run only one pass */
char filesok = 1;               /* =1 if ok to write to files */
char flogopen = 0;              /* =1 if log file is open */
char finsert = 1;               /* =1 to insert in devactive */
char do_numeric = 1;            /* =1 to use numeric sg devnames (sg0) */ 
char background = 0;            /* =1 if background, =0 if foreground app */
char moresn = 0;                /* =1 if 12-byte sernum, =0 if 8-byte sernum */
char fdevfs = 0;		/* =1 if DevFS is detected */
char fskipreset = 0;            /* =1 to skip any SCSI resets */
int interval   = 5;             /* sleep 5 sec interval between scans */ 
int sernum_len = 8;             /* usu 8, but some have 12 bytes */
int idev  = 0;
int ndev  = 0;
int nprocdev = 0;
int npart = 0;
int nrdy  = 0;
int nchgs = 0;
int lastnrdy = MAX_DEVS;
int iremove  = MAX_DEVS;
int iinsert  = MAX_DEVS;
SLOT_LIST slot[MAX_DEVS];
SLOT_LIST *pslot;
uchar  slotmode; 	/* SAF-TE: 3=#insertions, 4=status */
int slotoff = 0xff;	/* SAF-TE: offset into data for slot info */
int nslots = MAX_DEVS;	/* SAF-TE: number of slots */
int islot = 0;
char fsafte = 0; 	/* SAF-TE: is there a safte HSBP? */
char fses = 0;          /* SES: is there a SES HSBP? */ 
int max_try = 12;
char blank4[5]   = "    ";
char blank8[9]   = "        ";
char blank16[17] = "                ";
uchar inqcdb[6]     = { 0x12, 0, 0, 0, SGDATALEN, 0 };
uchar turcdb[6]     = { 0x00, 0, 0, 0, 0,    0 };
uchar startcdb[6]   = { 0x1B, 0, 0, 0, 0x01, 0 };
uchar defectcdb[10] = { 0x37, 0, 0x0d, 0, 0, 0, 0, 0x00, SGDATALEN, 0 };

/* Device Statuses for devlist.dstatus: */
char dstat_str[5][8] = { 
    "dead  ",  "failed", "notrdy", "ready ", "online"
};
DEVLISTX_TYPE devlist[MAX_DEVS];
struct {
    uchar vendor[9];
    uchar model[17];
    uchar fwrev[5];
    uchar sernum[13];
    uchar devtype; 
    uchar dstatus;      /* device status: DSTAT_* */
} devmatch[MAX_DEVS];
#define NDEVTYP  15
char *devtype_array[NDEVTYP] = {
    "Disk", /*0*/
    "Tape", /*1*/
    "Prtr", /*2*/
    "Proc", /*3*/
    "WORM", /*4*/
    "CDRM", /*5*/
    "Scan", /*6*/
    "Opti", /*7*/
    "MChg", /*8*/
    "Comm", /*9*/
    "Emul", /*10*/
    "Un11", /*11*/
    "RAID", /*12*/
    "Encl", /*13*/
    "Unkn"  /*14*/
};

/* List of vendors who use 12-byte serial numbers, instead of 8-byte. */
#define NVEND12  3
char *vendor12[NVEND12] = {"FUJITSU", "MAXTOR", "HP 73.4G" };

//////////////////////////////////////////////////////////////////////
// Local Subroutines
//////////////////////////////////////////////////////////////////////
void quit(int rc)
{
    if (flogopen) {
        fprintf(fdlog, "%s exiting rc = %d\n", progname, rc);
	fclose(fdlog);     /* close the log file */
    }
    flogopen = 0;
    fdlog = stderr;
    exit(rc);
} 

/*
 * showit
 * Outputs the message from buf to fdmsg, if not background,
 * and to fdlog, if opened ok.
 * Note that logit() takes var args and calls showit().
 *
 * Note that some of the fdebug output assumes foreground and 
 * uses printf directly.  So some debug output would go to 
 * /dev/null if background option is also set. 
 */
void showit(char *buf)
{
    if (filesok && !flogopen) { /* log is not open */
        fdlog = fopen(logfile, "a+");
        if (fdlog != NULL)
            flogopen = 1;
        else {
            printf("Cannot open %s, errno = %d\n", logfile, errno);
            fdlog = stderr;
        }
    }
    if (flogopen)
        fprintf(fdlog, "%s", buf);    /*write to log */
    if (!background)
        fprintf(fdmsg, "%s",  buf);    /*message display defaults to stdout */
}

/*
 * sdstat
 * Scan to see if this device is active or not.
 * This routine also checks if the devname is the last one in the 
 * system.  If so, we don't want to explicitly remove it.
 * Returns 0 if devname is active, 
 *         1 if failed, 
 *        -1 if missing
 * Sets *last (if not NULL) to:
 *     0 = not last, ok to remove
 *     1 = last one, do not remove
 *    -1 = error in getting status
 *    -2 = devname not found 
 */
int 
sdstat(char *devname, int *last)
{
    if (last != NULL) *last = 0;
    return(0);
}  /*end sdstat()*/

int 
rem_scsi_dev(int bus, int ch, int id, int lun)
{
	char cmd[40];
	FILE *fp;
        int n = 0;
	/* 
	 * echo "remove-single-device 0 0 3 0" >/proc/scsi/scsi
	 */
	fp = fopen("/proc/scsi/scsi","r+");
	if (fp == NULL) return(errno); 
	sprintf(cmd,"remove-single-device %d %d %d %d",bus,ch,id,lun);
	n = fprintf(fp,"%s", cmd);
	fclose(fp);
	strcat(cmd,"\n");
	showit(cmd);
        if (n <= 0) return(errno);
	return(0);
}

int 
add_scsi_dev(int bus, int ch, int id, int lun)
{
	char cmd[40];
	FILE *fp;
        int n = 0;

	/* 
	 * echo "add-single-device 0 0 3 0" >/proc/scsi/scsi
	 */
	fp = fopen("/proc/scsi/scsi","r+");
	if (fp == NULL) return(errno); 
	sprintf(cmd,"add-single-device %d %d %d %d",bus,ch,id,lun);
	n = fprintf(fp, "%s", cmd);
	fclose(fp);
	strcat(cmd,"\n");
	showit(cmd);
        if (n <= 0) return(errno);
	return(0);
}

char device_there(int idev)
{
   int sgfd;
   int ret;
   uchar sdbuf[SGDATALEN];
   int j;
   int devtype;

   /* if >= ndev, it didn't exist before. */
   if (idev >= ndev) return(0);  /* not there */

   if ((devlist[idev].devtype == DEV_PROC) ||
       (devlist[idev].devtype == DEV_ENCL))  /*Proc*/
	return(1);  /* proc, so report device is there */
   devtype = devlist[idev].devtype;

   /* assume device node is not already open */
   sgfd = open(devlist[idev].fname, O_RDWR | O_NONBLOCK);
   if (fdebug) printf("devthere [%d] %s open sgfd = %d\n",
			idev, devlist[idev].fname, sgfd);
   if (sgfd == -1) {
	return(0);  /* device node is not there */
   }

   /* See if device is ready */
   ret = sg_cmd(sgfd,turcdb,sizeof(turcdb),sdbuf,sizeof(sdbuf));
   if (fdebug) printf("devthere [%d] test_unit_ready ret = %d\n",idev,ret);
   if (ret < 0) {
	ret = sense_err(sgfd,sdbuf,sizeof(sdbuf),idev,ret,&j);
	if (fdebug) printf("devthere [%d] tur sense_err ret = %d\n",idev,ret);
   }
   close(sgfd);

   // if (ret >= 0) devlist[idev].dstatus = DSTAT_READY;
   // else devlist[idev].dstatus = DSTAT_NOTRDY;
   return(1);  /* device is there */
}  /* end device_there */

int
safte_inserted(int sgfd, int i, int hdev)
{
	char output[MSG_LEN];
	uchar sdbuf[SGDATALEN];
	int bus, ch, lun;
	int idev, j;
	int ret = 0;
        
	if (!fsafte) return(ret);
	/* bus, chn, lun are same as HSBP */
	bus = devlist[hdev].bus;
	ch  = devlist[hdev].chn;
	lun = devlist[hdev].lun;
	/* sgfd and hdev pertain to the HSBP device, not the disk. */
	/* Translate slot# (i) into idev of the disk */
	idev = ndev;
	for (j = 0; j < ndev; j++) {
	  if ((slot[i].id == devlist[j].tgt) &&
	      (bus == devlist[j].bus) && (ch == devlist[j].chn))
		 { idev = j; break; }
	}
	if (fdebug) showlog("slot %d slotid=%d, idev=%d devid=%d\n",
				i,slot[i].id,idev,devlist[idev].tgt);

	sprintf(output,"*** NEW SAF-TE INSERTION [%d] in slot %d, id %d\n",
	                        idev,i,slot[i].id);
	showit(output);
	if (idev == hdev) return(ret);

	/* If old scsi/aic driver, get data overrun here, see scsiras patch.*/
	ret = sg_cmd(sgfd,turcdb,sizeof(turcdb),sdbuf,sizeof(sdbuf));
	if (fdebug) showlog("hsbp test_unit_ready ret = %d\n",ret);

	if (!device_there(idev)) {
		/* If device node isn't already there, */
		/* do add-single-device now.  */
		ret = add_scsi_dev(bus,ch,slot[i].id,lun);
	        if (fdebug) 
		   showlog("add_scsi_dev[%d] ret = %d\n",slot[i].id,ret);
	}
        else if (!fskipreset) {
	        j = 2;  /* scsi bus reset */
	        ioctl(sgfd,SG_SCSI_RESET, &j);
	        showlog("safte_inserted: bus reset\n");
		sleep(1);  /* wait for reset to clear */
        }

	/* Do the actual sgevt insertion at the bottom */
	iinsert = idev;
	devlist[idev].dstatus = DSTAT_NOTRDY; /*not yet ready*/
	return(ret);
}

/* only call check_disk if device is a disk or Emul disk */
int check_disk(int idev, int sgfd)
{
   int ret, j;
   int gdl;
   int fretry;
   uchar gdbuf[SGDATALEN];

	   /* 
	    * If the device is a DISK:
	    * Do a get_defects command for sanity, since 
	    * sometimes an inquiry will return ok, but other
	    * commands will not.
	    */
	      ret = sg_cmd(sgfd,defectcdb,sizeof(defectcdb),
			  gdbuf,sizeof(gdbuf));
	      if (ret < 0) {  
		  j = sense_err(sgfd,gdbuf,sizeof(gdbuf),idev,ret,&fretry);
		  if (fretry) 
	               ret = sg_cmd(sgfd,defectcdb,sizeof(defectcdb),
			            gdbuf,sizeof(gdbuf));
	      }
	      if (ret < 0) {
		if (ret == SCHECK_CND) {
		    sprintf(output,"get_defects: sense err %02x:%02x:%02x\n",
			    gdbuf[2],gdbuf[12],gdbuf[13]);
		    showit(output);
		} else if (fdebug) {
		    sprintf(output,"get_defects ret = %d\n",ret);
		    showit(output);
		}
	      }
	      else { 
		 gdl = ((gdbuf[2] << 8) | gdbuf[3]) / 8;
	         if (fdebug) printf("grown defects=%d\n",gdl);
	      }
    if (ret >= 0) ret = 0;
    return(ret);
}

int check_proc(int idev, int sgfd)
{
	int ret;
	int i,j, fretry;
	uchar sdbuf[SGDATALEN];

        /* 
	 * If the device is a PROCESSOR (Hot-swap backplane),
	 * issue a SAF-TE command to get the insertion counts.
	 */
        if (fses) return(-1);
        /* new proc device if dcap/nslots == 0 */
        if (devlist[idev].dcap == 0) slotoff = 0xff;

	      if (slotoff == 0xff) {  /* initialize safte stuff */
		/* 0 - Read Enclosure Configuration */
	        ret = read_safte(sgfd, SAFTE_READ_ENCL_CONF, 
				 sdbuf, sizeof(sdbuf),idev);
		if (fdebug) showlog("read_safte/0 ret = %d\n",ret);
	        if (ret < 0) {
		  j = sense_err(sgfd,sdbuf,sizeof(sdbuf),idev,ret,&fretry);
		  if (fretry) 
	        	ret = read_safte(sgfd, SAFTE_READ_ENCL_CONF, 
					 sdbuf, sizeof(sdbuf),idev);
		}
	        if (ret != 0) { fsafte = 0; nslots = 0; }
	        else {   /* ret == 0 */
	         slotoff = sdbuf[0] + sdbuf[1];
	         nslots = sdbuf[2];
                 pslot = &slot[islot]; 
                 devlist[idev].dcap = nslots; /*save nslots for this proc*/ 
		 fsafte = 1;
	         if (fdebug) {
	            sprintf(output,
				"read_safte/0 ok, sgfd=%d off=%d nslots=%d\n",
				sgfd,slotoff,nslots);
		    showit(output);
		 }
	         for (i=0; i<nslots; i++) {
	             slot[i].ins = 0xFFFF;  /* init to max, fill in below */
                     slot[i].bus = devlist[idev].bus; 
                 }
		 /* 1 - Read Enclosure Status */
	         ret = read_safte(sgfd, SAFTE_READ_ENCL_STAT, 
					sdbuf, sizeof(sdbuf),idev);
		 if (ret != 0) {
		    j = sense_err(sgfd,sdbuf,sizeof(sdbuf),idev,ret,&fretry);
		    if (fretry) 
	         	ret = read_safte(sgfd, SAFTE_READ_ENCL_STAT, 
					sdbuf, sizeof(sdbuf),idev);
		 } 
	         if (fdebug) showlog("read_safte/1 ret = %d\n",ret);
		 if (ret == 0) {
	            for (i=0; i<nslots; i++)
	               pslot[i].id = sdbuf[slotoff + i];
	            if (fdebug) { 
			sprintf(output,
			       "read_safte/1 ok, offset=%d slotid[0,1]=%d,%d\n",
				slotoff,slot[0].id, slot[1].id);
			showit(output);
		    }
		    if (pslot[0].id >= pslot[1].id) { /*bogus HSBP correction*/
			pslot[0].id = 0;
			pslot[1].id = 1;
		    }
		    /* HSBP may not support Read_Device_Insertions, so try */
		    /* 4 - Read_Slot_Status.  */
		    slotmode = SAFTE_READ_DEV_STAT;
	            ret = read_safte(sgfd, slotmode, sdbuf,sizeof(sdbuf),idev);
	            if (fdebug) 
	                printf("read_safte/4 Read_Slot_Status ret = %d\n",ret);
		    if (ret == 0) {
			uchar *p;
			for (i=0; i<nslots; i++) {
				p = &sdbuf[4*i];
				pslot[i].stat = p[3];
			        if (fdebug) {
				   sprintf(output,
				       "slot %d, id %d: %02x %02x %02x %02x\n",
					i,pslot[i].id,p[0],p[1],p[2],p[3]);
				   showit(output);
				}
			}
		    } else {  /* error in Read_Slot_Status */
		        j = sense_err(sgfd,sdbuf,sizeof(sdbuf),idev,ret,
					&fretry);
			/* 3- Read_Device_Insertions */
			slotmode = SAFTE_READ_DEV_INS;
	                ret = read_safte(sgfd,slotmode,sdbuf,sizeof(sdbuf),
					 idev);
			if (fdebug) printf("read_safte/3 ret = %d\n",ret);
			if (ret == 0) {
			   slotoff = 0; 
	                   for (i=0; i<nslots; i++) {
	             		j = 2 * i;
				pslot[i].ins = sdbuf[j] + (sdbuf[j+1] << 8);
			   }
			}
		    }
	          }  /* safte/1 is ok */
                  islot += nslots;
	        } /*safte/0 is ok */
	      }  /*endif initial pass*/

	      else {  /* already initialized, so just get safte slot status */
		ret = read_safte(sgfd, slotmode, sdbuf, sizeof(sdbuf),idev); 
	        if (ret < 0) {
		  j = sense_err(sgfd,sdbuf,sizeof(sdbuf),idev,ret,&fretry);
		  if (fretry) 
	        	ret = read_safte(sgfd, slotmode, sdbuf, sizeof(sdbuf),idev);
		}
	      } /*end-else initialized*/

	      /* 
	       * Already have the slot status info, so
	       * fill in the slot array and analyze it. 
	       */
	      if (ret == 0) {
	         ushort tmpins;
	         uchar tmpstat;
	         if (fdebug) {
	           showlog("read_safte/%d ret ok\n",slotmode);
	           dump_log(sdbuf,sizeof(sdbuf),"safte_buf",0);
	           }
                 /* find pslot and nslots for this proc */
                 for (i=0; i<islot; i++)
                   if (slot[i].bus == devlist[idev].bus) break;
                 if (i >= islot) i = 0;
                 pslot = &slot[i];
                 nslots = devlist[idev].dcap;
	         /* check for an insertion event in each slot */
		 if (slotmode == 4) {
	           for (i=0; i<nslots; i++) {
	             j = 4 * i;
	             tmpstat = sdbuf[j+3];
		     if (fdebug) {
			logit("slot=%d, id=%d, %02x %02x %02x %02x stat=%x\n",
	                        i, pslot[i].id, sdbuf[j],sdbuf[j+1],
				sdbuf[j+2], tmpstat, pslot[i].stat);
		     }
		     if (tmpstat != pslot[i].stat) {  /* a change happened */
			/* if it was removed, and is now inserted, act. */
			if ((pslot[i].stat & 0x01) == 0 && (tmpstat & 0x01)) 
				ret = safte_inserted(sgfd,i,idev);
			pslot[i].stat = tmpstat;
		     }
		   }
		 } else {  /* slotmode == 3 */
	           for (i=0; i<nslots; i++) {
	             j = 2 * i;
	             tmpins = sdbuf[j] + (sdbuf[j+1] << 8);
		     if (fdebug) {
	                if (tmpins > 0) {
	                   logit("slot %d, id %d had %d insertions\n",
	                         i,pslot[i].id,tmpins);
		        }
		     }
	             if (tmpins > pslot[i].ins) 
			ret = safte_inserted(sgfd,i,idev);
	             pslot[i].ins = tmpins;
	           }
	         }
	      } else {  /* read_safte error */
	         if (ret == SCHECK_CND) { 
	              sprintf(output,
				"check_proc: read_safte sense = %x/%x/%x\n",
				sdbuf[2], sdbuf[12], sdbuf[13]);
	         } else
	              sprintf(output,"check_proc ret = %d\n", ret);
	         showit(output);
	      }
	return(ret);
}  /*end check_proc*/

/* 
 * scandev
 * This function sends inquiries to each possible device and
 * collects the information needed to display the active devices.
 */
static int
scandev(int bmode)
{
   int i, j, idev;
   char fname[64];
   char sdname[64];
   char tmpname[64];
   uchar dbuf[SGDATALEN];
   int sgfd;
   int num_err;
   int ret = 0;
   int iret;
   DEVFS_LIST devfs_components = {0,0,0,0,0};

   /*
    * Iterate through the list of possible devices by number.
    */
   ndev = 0;
   nprocdev = 0;
   num_err = 0;
   for (idev = 0; (idev < MAX_DEVS) && (num_err < MAX_ERRS); idev++)
   {
	/* Open each corresponding sg device name */
	make_dev_name(idev,fname,sdname,do_numeric, &devfs_components);
	if (devfs_components.all_leaves_searched != 0) {
		devfs_components.all_leaves_searched = 0;
		break;
	}
	if (nprocdev > 0) {
	   /* Proc devices don't use an sdname, so get 
	    * the previous block device name that will match. */
	   make_dev_name((idev-nprocdev),tmpname,sdname,do_numeric, 
			 &devfs_components);
	}
	devlist[idev].dstatus = DSTAT_DEAD;  /*initial value*/
	sgfd = open(fname, O_RDWR | O_NONBLOCK);
	if (sgfd < 0) {
	   if (fdebug) {
		if (errno == ENXIO) showlog("%s no device\n",fname);
		else showlog("%s open errno: %d\n",fname,errno);
	   }
	   if (errno == EBUSY) continue;
	   else if (errno == ENODEV || errno == ENOENT 
		    || errno == ENXIO) {
                        ++num_err;
                        continue;
	   }
	   else {  /* EACCES or other, needs attention */
		if (fdebug)
			showlog("Error %d opening %s\n",errno,fname);
                        ++num_err;
                        continue;           
	   }
	} else {  /* opened ok, the device exists (or did exist) */
	   struct { 
	       int dev_id;
	       int host_unique_id; 
	       } idlun;
	   int bus,ch,id,lun;
	   int fretry;
	   int f, ilen, slen;
	   int devtype = 0;
	   int emul = 0;
           int fLiveDisk;

	   strcpy(devlist[idev].fname,fname);
	   ret = ioctl(sgfd, SCSI_IOCTL_GET_BUS_NUMBER, &bus);
	   if (ret < 0) ret = SIOCTL_ERR;
	   if (ret >= 0) 
	 	ret = ioctl(sgfd, SCSI_IOCTL_GET_IDLUN, &idlun);
	   if (ret >= 0) 
	 	ret = ioctl(sgfd, SG_EMULATED_HOST, &emul);
	   if (ret < 0) { 
		ret = SIOCTL_ERR; 
		goto abortsg; 
	   } 
	   ch  = (idlun.dev_id >> 16) & 0xff;
	   id  = idlun.dev_id & 0xff;
	   lun = (idlun.dev_id >> 8) & 0xff;
	   if (bmode == 1) {
	      f = fcntl(sgfd, F_GETFL);
	      f = f & (~O_NONBLOCK);  /* set to blocking mode */
	      ret = fcntl(sgfd, F_SETFL, f);      
	   }
	   devlist[idev].bus = bus;
	   devlist[idev].chn = ch;
	   devlist[idev].tgt = id;
	   devlist[idev].lun = lun;

           if (!fdevfs) {
              /* find the real sdname, if sysfs is there. */
              j = get_dev_maj_min(bus,ch,id,lun,
                                &devlist[idev].maj,&devlist[idev].min);
              if (fdebug) printf("[%d] jret=%d maj=%d, min=%d\n",idev,j,
                                devlist[idev].maj,devlist[idev].min);
              if (j == 0) {
                j = find_mmdev(devlist[idev].maj,devlist[idev].min,sdname);
                if (fdebug) printf("[%d] jret=%d sdname = %s\n",idev,j,sdname);
              } else {
                j = find_blockdev(bus,ch,id,lun,sdname);
                if (fdebug) printf("[%d] bdev ret=%d sdname = %s\n",
					idev,j,sdname);
	      }
           }
	   if (sdname[0] == 0) strcpy(devlist[idev].sdname,"    ");
	   else strcpy(devlist[idev].sdname,sdname);
	   /* 
	    * Do a SCSI inquiry and save the info. 
	    */
	   ilen = sizeof(dbuf)-1;
           memset(dbuf,0,sizeof(dbuf));
	   ret = sg_cmd(sgfd,inqcdb,sizeof(inqcdb),dbuf,sizeof(dbuf));
	   if (fdebug) printf("inquiry ret=%d, errno=%d\n", ret,errno);
	   if (ret < 0) {   
		f = sense_err(sgfd,dbuf,sizeof(dbuf),idev,ret,&fretry);
		if (fretry) {
                   memset(dbuf,0,sizeof(dbuf));
	           ret = sg_cmd(sgfd,inqcdb,sizeof(inqcdb),
			        dbuf,sizeof(dbuf));
                }
	   }  /*endif sense err*/

           if (ret >= 0) iret = 0;
           else iret = ret;
           devtype = dbuf[0] & 0x1f;
           if (iret == 0) {
              int fbad;
              uchar ci;
              int ftrunc = 1;    /*ARC*/
              ilen = dbuf[4];
              fbad = 0;
              if (ilen > 36 && ilen < 44) ilen = 46; /*Fix Fujitsu fw bug*/
              if (ilen < 36 && ilen < ret) ilen = ret;
              if (ilen >= sizeof(dbuf)) ilen = sizeof(dbuf)-1;
              if ((ilen < 24) || (devtype > DEVT_MAX)) {
                 j = 32; fbad = 1;
              } else for (j = 8; j < ilen; j++) {
                 ci = dbuf[j];
                 if (ci > 0x7f) {
                    if (fdebug)
                       showlog(
                           "[%d] bad inquiry char 0x%02x (%c), offset=%d\n",
                                idev, ci, ci, j);
                    if (ftrunc && j > 24) {   /* truncate it with no error */
                       dbuf[j] = 0;
                       ilen = j;
                       break;
                    } else {        /* have bad inquiry data */
                       fbad = 1;
                       break;
                    }
                }
              }
              if (fbad) {
                 /* gets here if standby (not-active) disk is removed */
                 if (fdebug)
                     showlog("Inquiry error: len=%d type=%d badc=%02x\n",
                                        ilen,devtype,dbuf[j]);
                 iret = SIZE_ERR;
                 ilen = j; /*truncate, was 24*/
              }
	      dbuf[ilen] = 0; /*stringify*/
	      if (fdebug)
		dump_log(dbuf,ilen,"inq buffer",0);
	   } else if (fdebug && iret == SCHECK_CND) {
	 	sprintf(output,"inq: scsi sense err %02x:%02x:%02x\n",
			   dbuf[2],dbuf[12],dbuf[13]);
		showit(output);
	   }

         ret = iret;
	 if (iret == 0) {  /* only check more if inquiry was ok */
             if (emul) { devtype = DEV_EMUL; } /*Emul = 10., usu usb-storage*/
	     if (devtype >= NDEVTYP) devtype = NDEVTYP - 1;  /*Unkn*/
             if (devtype == DEV_PROC || devtype == DEV_ENCL) {
                 /* proc, HotSwap backplane */
                 devlist[idev].devtype = devtype;
                 fillstr(devlist[idev].vendor,&dbuf[8],8);
                 ret = check_proc(idev,sgfd);
                 if (ret != 0) {   /* fsafte==0 */
                    if ((dbuf[6] & ENCSERV) || (devtype == DEV_ENCL))
                       ret = check_ses(idev,sgfd);
                 }
             } else if (devtype == DEV_DISK) {          /* disk */
                 ret = check_disk(idev,sgfd);
             }
         }
	 if (fdebug) {
		/* display the device */
		sprintf(output,"dev[%d] iret=%d, ret=%d, len=%d %d:%d:%d:%d %s %s %s\n",
		       idev,iret,ret,ilen,bus,ch,id,lun,fname,sdname,&dbuf[8]);
		showit(output);
	 }
	 /* 
	  * Now update the device list info, if valid 
	  */
	 if (iret < 0) {
		devlist[idev].ret = iret;
		devlist[idev].dstatus = DSTAT_FAIL;  /*it is failed*/
	        devlist[idev].sernum[0] = 0;
	 } else {    // If here, valid, so update the devlist
	        devlist[idev].fwrev[0] = 0;
		dbuf[ilen] = 0;  /* end of buffer */
		devlist[idev].devtype = devtype;
		devlist[idev].ret = ret;
		if (ret == 0) devlist[idev].dstatus = DSTAT_READY;  /*ready*/
                else if (ret == SCHECK_CND) 
			devlist[idev].dstatus = DSTAT_NOTRDY; /*not yet ready*/
                else 	devlist[idev].dstatus = DSTAT_FAIL; 
		if (iret == 0) {
		  fillstr(devlist[idev].vendor,&dbuf[8],8);
		  fillstr(devlist[idev].model,&dbuf[16],16);
		  fillstr(devlist[idev].fwrev,&dbuf[32],4);
		  slen = sernum_len;  /*set default from global value*/
		  for (i=0; i < NVEND12; i++) {
		    if (strncmp(devlist[idev].vendor,vendor12[i],
				strlen(vendor12[i])) == 0)
			 slen = 12;
		  }
		  if (dbuf[44] <= ' ') slen = 8;
		  if ((ilen > 44) && (dbuf[36+8] > ' ') && 
                      (devtype == DEV_DISK)) 
			 slen = 12; /*apparently a long sernum disk*/
		  strncpy(devlist[idev].sernum,&dbuf[36],slen);
		  devlist[idev].sernum[slen] = 0;
		  /* could validate sernum here */
		  if (ilen >= 54) {
		    strncpy(devlist[idev].other,&dbuf[44],10);
		    devlist[idev].other[10] = 0;
	          }
		} /*endif iret==0*/
	 } 
	 fLiveDisk = 0;
         if (iret == 0) {   /* status ok, so live */
	     if (devtype == DEV_DISK) /*it is a live Disk*/ fLiveDisk = 1;
	     else if (IsSATA(idev)) fLiveDisk = 1; 
	 }
	 /* check the status of this disk in the system */
	 if (fLiveDisk) {
		/* is it active? */
		devlist[idev].sdactive = sdstat(devlist[idev].sdname,NULL);
		if (fdebug) 
		   showlog("dev[%d] sdstat = %d\n",idev,
			       devlist[idev].sdactive);
	 } else {  /* either inquiry error, or not a disk */
		devlist[idev].sdactive = -1;
	 }
         abortsg: 
	    ndev = idev + 1;
	    if ((devtype == DEV_PROC) || (devtype == DEV_ENCL)) nprocdev++;
	    close(sgfd);
	    if (ret != 0) {
		showlog("scandev: device %d error=%d, errno=%d\n", 
			idev,ret,errno);
	    }
	} /*endif good open*/
   }  /*end for each device*/

   /* set up the next fname */
   strcpy(devlist[ndev].fname,fname);
   return ret; 
}  /* end scandev */

/* 
 * mkdaemon
 * Called when sgdiskmon needs to become a daemon/background process.
 */
int 
mkdaemon(int fchdir, int fclose)
{
    int fdlimit = sysconf(_SC_OPEN_MAX);
    int fd = 0;
 
    switch (fork()) {
        case 0:  break;
        case -1: return -1;
        default: _exit(0);          /* exit the original process */
    }
    if (setsid() < 0) return -1;    /* shouldn't fail */
    switch (fork()) {
        case 0:  break;
        case -1: return -1;
        default: _exit(0);
    }
    if (fchdir) chdir("/"); 
    if (fclose) {
	/* Close stdin,stdout,stderr and replace them with /dev/null */
	while (fd < fdlimit) close(fd++);
	open("/dev/null",O_RDWR);
	dup(0); dup(0);
    }
    return 0;
}

/* 
 * do_insert
 * Handle hot-inserted disks.
 */
int do_insert(int iinsert)
{
	int ret = 0;
	int i, j, c;
	int jslot = MAX_DEVS;
        char *sddev;

	ret = 0;
	if (iinsert != MAX_DEVS) {
	   if (fdebug)
	      showlog("do_insert/pre: i=%d dstat=%d ret=%d type=%d\n",
			iinsert,devlist[iinsert].dstatus,devlist[iinsert].ret,
			devlist[iinsert].devtype);
	   if ((devlist[iinsert].dstatus <= DSTAT_FAIL) || 
	       (devlist[iinsert].devtype == DEV_PROC) || 
               (devlist[iinsert].devtype == DEV_ENCL)) 
                  iinsert = MAX_DEVS; 
	}
	if (iinsert == MAX_DEVS) {  /*iinsert not set*/
	     /* find which one was inserted */
             for (i = 0; i < ndev; i++) {  
		/* See if this disk position came ready again */
		if (devlist[i].dstatus > devmatch[i].dstatus) {
			/* This one was dead and is back alive again */
		   	if (fdebug) 
			  showlog(
			       "do_insert/st: i=%d dstat=%d ret=%d sdstat=%d\n",
				i,devlist[i].dstatus,devlist[i].ret,
				devlist[i].sdactive);
			iinsert = i; 
			break;  /* found it, done */
		     }
		/* See if this is a new serial number in devlist */
		if (devlist[i].sernum[0] != 0) {
		   for (j = 0; j < ndev; j++) {
		      if (strcmp(devlist[i].sernum,devmatch[j].sernum)==0) 
		   	 break;  /*matches, so not new*/
		   }
	           if (j >= ndev) {  /*wasn't in the list before*/
		      if (fdebug) 
			showlog(
			       "do_insert/sn: i=%d dstat=%d ret=%d sdstat=%d\n",
				i,devlist[i].dstatus,devlist[i].ret,
				devlist[i].sdactive);
		      iinsert = i; 
		      break; 
		   }
		}
	     } /*end for i*/
	}  /*endif iinsert not set */
	/* failsafe check for wrong iinsert */
	if ((devlist[iinsert].dstatus <= DSTAT_FAIL) || 
	    (devlist[iinsert].devtype == DEV_PROC) || 
            (devlist[iinsert].devtype == DEV_ENCL)) 
                  iinsert = MAX_DEVS; 

	if (iinsert != MAX_DEVS) {
              sprintf(output,"insert: disk %d was inserted.\n",iinsert);
              showit(output);
	      sddev = devlist[iinsert].sdname;
              jslot = devlist[iinsert].tgt;
	      if (fsafte) {
	        /* if SAFTE, write slot status */
	        ret = write_slots(iinsert,0x01);
	        j = jslot;
	        ret = 0;
	      }
#ifdef OLD
              if (devlist[iinsert].sdactive == 0) {
		 /* We only get sdactive==0 if all partitions show active
		  * in /proc/sdstat. */
		 logit("%s is already active in the system\n",
			devlist[iinsert].sdname);
	      } else
#endif
	      /* Loop for each partition on the device, npart==1 */
	      for (j = 1; j <= npart && ret >= 0; j++) {
	         /* alert sgevt with the inserted disk */
		 {   		// if (ret == 0) 
                   /* use sgevt to do the systemhotadd */
                   sprintf(output,"%s %s %s %d %d",sgevtcmd,"Insert",sddev,
				jslot,iinsert); /*was iinsert,j); */
                   c = system(output);  /* execute sgevt command */
                   logit("%s\ninsert: system ret = %d\n", output,c);    
		   /* set dstatus on next pass, not here. */
		 }
	      }
	}
	return(ret);
} /*end do_insert*/

/*
 * do_remove
 * Handle hot-removed disks.
 */
int do_remove(int iremove)
{
	int ret = 0;
	int i, j, k;
        int c = 0;
	char *sddev;

	if (iremove == MAX_DEVS) 
              for (i = 0; i < ndev; i++) {
		if (devlist[i].devtype == DEV_DISK &&   /*Disk*/
		    (devlist[i].dstatus <= DSTAT_FAIL))
			{ iremove = i; break; } 
		if (IsSATA(i) &&
		    (devlist[i].dstatus <= DSTAT_FAIL))
			{ iremove = i; break; }  
	      }
        if (devlist[iremove].devtype != DEV_DISK && !IsSATA(iremove)) { 
               /* specified, but not a Disk, and not SATA */
               sprintf(output,"remove(%d): %s is not a disk\n", iremove,
			devlist[iremove].sdname);
	       showit(output);
               /* if HSBP/proc, will need a reset on next pass */
	} else {    /* (iremove < ndev) */
	       int jslot = MAX_DEVS;
	       i = iremove;
	       if (iremove >= ndev) {
		   i = ndev;
                   logit("remove: disk %d removed already\n",i);
               } 
               sddev = devlist[i].sdname;
               jslot = devlist[i].tgt;  /* SCSI target ID == slot */
               logit("remove: sdname = %s, dstat = %d, idev = %d npart=%d\n",
	  	       devlist[i].sdname,devlist[i].dstatus,i,npart);
	       if (fsafte) {
	         /* if SAFTE, write slot status */
	         ret = write_slots(i,0x02);
	         j = jslot;
	         ret = 0;  // continue even if write_slots error 
	       }
	       /* Loop for each partition on the device, npart==1 */
	       for (j = 1; j <= npart && ret >= 0; j++) {
		 {   		/* (ret == 0)  do sgevt */
		    k = sdstat("part",&c);
		    if (c == 1) {
                      logit("%s is last in the system, do not remove.\n",sddev);
		      /* need to leave it online, so not try to re-add next time */
		      devlist[i].dstatus = DSTAT_ONL;
		      ret = -3;  /* signal to increment nrdy */
		    } else {   		//  do sgevt
		      sprintf(output,"%s %s %s %d",sgevtcmd,"Fail",sddev,jslot);
		      ret = system(output);  /* execute sgevt command */
		      logit("%s\nremove: system ret = %d\n", output,ret);    
		    }
		 }  /*end-else ret==0*/
	       }  /*end for*/
	}  /*endif valid iremove*/
	return(ret);
} /*end do_remove*/

/* 
 * check_devstatus
 * This routine checks each device to see what status it really has.
 * It tracks the previous set of information in devmatch and uses
 * the serial number for uniqueness.
 */
void check_devstatus(void)
{
	int i, j, k; 
        int fLiveDisk;
        nrdy = 0;
	for (i = 0; i < ndev; i++) {
	   if (lastnrdy != MAX_DEVS) {  /*not first time*/
	     if (strcmp(devlist[i].sernum,devmatch[i].sernum)!=0) {
		/* 
		 * Serial number doesn't match what it was before.
		 * This mismatch could either be an impostor or a real 
		 * replacement.  Assume that it is a real replacement 
		 * unless the serial number is missing.
		 * I saw one impostor return all blanks for sernum, 
		 * but some drives may have leading blanks on a valid
		 * sernum of 4 digits.  If all blanks, invalid.
		 */
#ifdef NOT_USED
		/* can have SW/HW RAID disks with no sernum, so skip this */
		if (devlist[i].sernum[0] == 0) devlist[i].dstatus = DSTAT_FAIL;
		for (j=0;j<sernum_len;j++)  /* sernum missing, all spaces? */
			if (devlist[i].sernum[j] != 0x20) break;
		if (j == sernum_len) devlist[i].dstatus = DSTAT_FAIL;
#endif
		if (fdebug) {
	          sprintf(output,"disk %d sernum changed, sn0=%x oldsn0=%x dstat=%d iremove=%d\n",
			  i,devlist[i].sernum[0],devmatch[i].sernum[0],
			  devlist[i].dstatus,iremove);
	          showit(output);
		}
		if (devmatch[i].sernum[0] == 0 && devlist[i].sernum[0] != 0) { 
		    /* prev was failed, now chg */
		    if (devlist[i].ret == 0 && 
			devlist[i].dstatus == DSTAT_FAIL) {
			/*good inquiry now, must be inserted*/
			devlist[i].dstatus = DSTAT_READY; /*now ok*/
			iinsert = i;
		    }
		}
		if (devlist[i].dstatus == DSTAT_FAIL) {
			/* found an impostor, failed */
			/*devlist[i].devtype = 0; * assume failed type=disk */
			devlist[i].devtype = devmatch[i].devtype; /*prev type*/
			iremove = i;
		}
	     } /*endif mismatch*/
	   } /*endif not first time*/
	   /*
	    * Check if a disk is shown with a different serial number than 
	    * before.  If so, it is probably removed, and one of the other
	    * disks is answering in its place (duplicate).  Mark it as dead. 
	    * Note that this assumes that you won't remove an active disk
	    * and then re-insert the same one with a different SCSI ID.
	    */
           if (!frunonce)
	   for (j = 0; j < ndev; j++) {
		/* 
		 * Check if there is another device listed with same sernum.
		 * Non-disks can also appear here as impostors, and 
		 * those serial numbers are not always globally unique, 
		 * so also match by model. 
		 */
		k = strcmp(devlist[i].model,devlist[j].model);
		if (k==0 && strcmp(devlist[i].sernum,devlist[j].sernum)==0) {
		   if (j != i) { /*two different devices match sernum*/
		        int idup;
		      /*
		       * If same serial number appears twice, mark the one
		       * that is out of place as dead.  Use sdstat to find 
		       * out which one is the impostor.
		       */
			idup = MAX_DEVS;
			if (fdebug) showlog(
				"dup_chk: [%d] ret=%d sdstat=%d [%d] ret=%d sdstat=%d\n", 
				i,devlist[i].ret,devlist[i].sdactive, 
				j,devlist[j].ret,devlist[j].sdactive);
			if ( (devlist[j].sdactive != 0)) idup = j;
			else if ( (devlist[i].sdactive != 0)) idup = i;
	        	else if(strcmp(devlist[j].sernum,devmatch[j].sernum)!=0)
			      idup = j;   /*j sernum is different*/
	        	else if(strcmp(devlist[i].sernum,devmatch[i].sernum)!=0)
			      idup = i;   /*i sernum is different*/
			else if (devlist[j].ret != 0)   /*bad inquiry*/
			      idup = j; 
			else if (devlist[i].ret != 0)  /*bad inquiry*/
			      idup = i;
			if (idup != MAX_DEVS) {
			   /* idup is the impostor, mark it dead */
			   devlist[idup].dstatus = DSTAT_FAIL;
	        	   strcpy(devlist[idup].vendor,devmatch[idup].vendor);
	        	   strcpy(devlist[idup].model,devmatch[idup].model);
	        	   strcpy(devlist[idup].fwrev,devmatch[idup].fwrev);
	        	   strcpy(devlist[idup].sernum,devmatch[idup].sernum);
			   devlist[idup].devtype = devmatch[idup].devtype; 
			   iremove = idup;
			   if (fdebug) {
			      sprintf(output,"device %d is duplicate\n",idup);
			      showit(output);
			   }
			}  /*endif found impostor*/
		   }  /*endif j!=i*/
		}  /*endif sernums match*/
	   } /*end for j*/
	   fLiveDisk = 0;
	   if (devlist[i].devtype == DEV_DISK) fLiveDisk = 1;
	   else if (IsSATA(i)) fLiveDisk = 1;
	   /*
	    * Check this disk's status in the system.
	    */
	   if (fLiveDisk) {		/* is a disk */
		k = devlist[i].sdactive; 
		if (k == 0 && 			  	/* active in system */
		    devlist[i].dstatus >= DSTAT_READY)  /* and ready */
		    devlist[i].dstatus = DSTAT_ONL;     /* mark it online */
		else if (k != 0 && 
			 devlist[i].dstatus == DSTAT_READY) {
		     /* ready, configured (c), but not active (k) in system */
		     iinsert = i;  /* mark it to be inserted */
		     if (devlist[i].dstatus == devmatch[i].dstatus) nchgs++;
		} else if (k == 1) { 			/* failed in system */
			/* decided above if it is failed, or newly inserted */
			if (fdebug) {
			   sprintf(output,
				"dev=%d %s shows faulty status=%x ret=%d\n",
				i,devlist[i].sdname,devlist[i].dstatus,
				devlist[i].ret);
			   showit(output);
			}
		}
	     
	      if (lastnrdy == MAX_DEVS) { /*first time*/
		/* 
		 * Check if this disk has become ready since the last
		 * reboot and should be inserted as part of the system.  
		 * This is important if the system booted up degraded.
		 */
		if (fdebug) showlog("1st: dev %d dstat=%d sdactive=%d\n",i,devlist[i].dstatus,k);
		if (devlist[i].dstatus >= DSTAT_READY) {  /* disk is ready*/
		   if (k != 0) {
		     /* ready, configured (c), but not active (k) in system */
		     nchgs++;
		     iinsert = i;  /* mark it to be inserted */
		   }
		}
	      } /*endif first time*/
	   }  /*endif is a disk*/

	   if (lastnrdy != MAX_DEVS && 
	       devlist[i].dstatus != devmatch[i].dstatus) nchgs++;
           /* ndev may include some that are offline or dead, so count nrdy. */
           if (devlist[i].dstatus > DSTAT_NOTRDY) nrdy++;
	}  /*end check devices loop*/
	return;
}  /*end check_devstatus*/

int
main(int argc, char **argv)
{
   int ret;
   int i, c;
   ulong ltime1;
   int bmode = 1;
   char *pfwrev;
   char *psernum;
   char *pvendor;
   char *pmodel;
   char *psdname;

   fdmsg = stdout;
   sprintf(output2,"\n     %s utility v%s for SCSI Disk monitoring\n",
		progname, progver);

   while ((c = getopt(argc, argv, "ebmnrst:x?")) != EOF) {
        switch (c) {
        case 'x':
            fdebug = 1;
            break;
        case 'b':
            background = 1;
            break;
        case 's':
            fskipreset = 1;
            break;
        case 'm':
            moresn = 1;
	    sernum_len = 12;
            break;
        case 'e':
            filesok = 0;
            break;
        case 'r':
            frunonce = 1;
            break;
        case 'n':
            do_numeric = 0;
            break; 
        case 't':
            interval = atoi(optarg);
            break;
	default:
            printf("%s", output2);    /*prog header*/
            printf("Usage: %s [-bemnrx -t time]\n",progname);
	    printf(" -b  Run in Background as a daemon\n");
	    printf(" -e  Do not write to any files.\n");
	    printf(" -m  Max serial number length of 12 instead of 8 bytes\n");
	    printf(" -n  Turn off Numeric device names, use alpha instead\n");
	    printf(" -r  Run one scan, check status, but do not loop\n");
	    printf(" -s  Skip any SCSI Resets for error recovery.\n");
	    printf(" -t  Time in seconds between each scan, default 5\n");
	    printf(" -x  Show eXtra debug messages\n");
            exit(1); 
	}
   }
   if (!background) printf("%s", output2);  /*prog header*/

   /* only run this as superuser */
   i = geteuid();
   if (i > 1) {
        printf("%s: Not superuser (%d)\n", progname,i);
        exit(1);
   } 

   /* convert to a daemon if background */
   if (background) {
	ret = mkdaemon(1,1);
        if (ret != 0) {
           printf("%s: Cannot become daemon, ret = %d\n", progname,ret);
           exit(1);
	}
   }

   /* open log file */
   if (filesok)
        fdlog = fopen(logfile, "a+");
   if (fdlog == NULL) {
        flogopen = 0;
        filesok = 0;
        fdlog = stderr;
   } else {                    /* logfile open successful */
        flogopen = 1;
        time((time_t *) & ltime1);
        showlog( "\n--- %s v%s log started at %s", progname, progver,
                ctime((long *)&ltime1));   // ctime outputs a '\n' at end
	showlog(output2);  /* prog header */
   }
   memset(devlist,0,sizeof(devlist));
   npart = 1;   /* get number of system partitions */
   if (fdebug) printf("main: npart=%d\n",npart);
   if (npart == 0) showlog( "WARNING:  No system partitions found in /etc/raidtab\n");
   if (npart == -1) npart = 0;  /*mismatched partitions, so don't manage them*/
   while (1) 
   {
        nchgs = 0;
	/* 
	 * scandev gets the scsi data from the devices, 
	 * sets ndev, and fills devlist.
	 */
	ret = scandev(bmode);
	if (fdebug || ndev == 0) printf("scan ret = %d\n",ret);
	bmode = 0;

	/*
	 * check_devstatus scans the devlist for unique serial
	 * numbers and matches against the previous snapshot.
	 */
	check_devstatus();

	/*
	 * Now, if there have been any changes since the previous snapshot,
	 * display the new list of devices.
	 */
	if (fdebug) {
           sprintf(output,"nchgs=%d nrdy=%d lastnrdy=%d ndev=%d\n",
			nchgs,nrdy,lastnrdy,ndev);
	   showit(output);
	}
	/* First time, insert dev if iinsert is set */
	if ((lastnrdy == MAX_DEVS && iinsert != MAX_DEVS) &&
	    (devlist[iinsert].ret == 0)) {
		/* iinsert is set, and that device is alive. */
		if (fdebug) {
	          sprintf(output,"Need to insert %s\n",devlist[iinsert].sdname);
	          showit(output);
		}
		lastnrdy = nrdy - 1; 
	}
	if (lastnrdy == MAX_DEVS || nchgs > 0 || lastnrdy != nrdy) {
           /* Something changed.  Display header & devices */
	   sprintf(output,
	    "Num Name      bus:ch:id:lun Type Vendor   Device_Model     FRev Serial#  Status\n");
	   showit(output);
	   for (i = 0; i < ndev; i++) {
	      char *tmpl;
              if (devlist[i].dstatus == DSTAT_DEAD) continue;
	      if (devlist[i].vendor[0] == 0) pvendor = blank8;
	      else pvendor = devlist[i].vendor;
	      if (devlist[i].model[0] == 0) pmodel = blank16;
	      else pmodel = devlist[i].model;
	      if (devlist[i].fwrev[0] == 0) pfwrev = blank4;
	      else pfwrev = devlist[i].fwrev;
	      if (devlist[i].sernum[0] == 0) psernum = blank8;
	      else psernum = devlist[i].sernum;
	      if ((devlist[i].devtype == DEV_PROC) ||
	          (devlist[i].devtype == DEV_ENCL)) psdname = blank8;
	      else psdname = devlist[i].sdname;
	      /* Display each device */
	      if (fdevfs) 
		 tmpl = "%d %s\n  %s\n\t\t    %d:%d:%d:%d %s %s %s %s %s %s\n";
	      else tmpl = "%d %s %s %d:%d:%d:%d %s %s %s %s %s %s\n";
	      /* TODO: if (devlist[i].fses) show something */
	      sprintf(output,tmpl, i,
	              devlist[i].fname, psdname, devlist[i].bus,
		      devlist[i].chn, devlist[i].tgt, devlist[i].lun,
	       	      devtype_array[devlist[i].devtype],
		      pvendor, pmodel, pfwrev, psernum, 
		      dstat_str[devlist[i].dstatus]);
	      showit(output);
	   }
	}

	/* 
	 * Decide if there is any action to take.
	 * If first time, no action unless insert needed.
	 * If hot-removal, invoke do_remove for sgevt Fail.
	 * If hot-insert, invoke do_insert for sgevt Insert.
	 */
        if (lastnrdy == MAX_DEVS) {     /* first time */
            sprintf(output,"monitoring %d scsi devices ...\n", nrdy);
	    showit(output);
	    if (nrdy == 0) {
		sprintf(output,"Cant open any sg scsi devices, errno = %d\n",
		        errno);
		if (errno == ENODEV) strcat(output,"Make sure CONFIG_CHR_DEV_SG is set in /usr/src/linux/.config\n");
		else if (errno == ENOENT) strcat(output,"Check /dev/sg* device names\n");
		strcat(output,"Try 'modprobe sg' if not already loaded\n");
	        showit(output);
		quit(1);
	    }
        } else if (nrdy > lastnrdy) {   /* disks were added/inserted */
	    sprintf(output,"***** %d->%d: disk inserted (%d) ******\n", 
		   lastnrdy, nrdy, iinsert);
	    showit(output);
	    ret = do_insert(iinsert);
            iinsert = MAX_DEVS;  /*done, so clear iinsert*/

        } else if (nrdy < lastnrdy) {     /* disks were removed */
            sprintf(output,"***** %d->%d: disk removed (%d) ******\n", 
		    lastnrdy, nrdy, iremove);  
	    showit(output);
	    ret = do_remove(iremove);
	    // if (ret == -3) nrdy++;  /* last one, so add it back as ready */
	    iremove = MAX_DEVS;

        } else if (lastnrdy != MAX_DEVS && nchgs > 0 && nrdy == lastnrdy) {
	    /* One or more changes occurred (i.e. remove & insert), 
	     * but the same number of devices are ready.
	     * The iremove & iinsert should already be set from above.
	     * Do any removes first, if valid.
	     */
            sprintf(output,"***** %d->%d: %d changes occurred (%d/%d) ******\n",
		    lastnrdy, nrdy, nchgs, iremove,iinsert);  
	    showit(output);
	    ret = do_remove(iremove);
	    // if (ret == -3) nrdy++;  /* last one, so add it back as ready */
	    iremove = MAX_DEVS;
	    ret = do_insert(iinsert);
            iinsert = MAX_DEVS;     /*done, so clear iinsert*/
	}

	/* 
	 * Save the current good snapshot for the next pass.
	 */
	lastnrdy = nrdy;
        for (i = 0; i < ndev; i++) {  
	   if (devlist[i].dstatus > DSTAT_FAIL) {
	      if (fdebug && devlist[i].dstatus != DSTAT_ONL) 
		 printf("dev %d stat=%d, saving\n",i,devlist[i].dstatus);
	      strcpy(devmatch[i].vendor,devlist[i].vendor);
	      strcpy(devmatch[i].model, devlist[i].model);
	      strcpy(devmatch[i].fwrev, devlist[i].fwrev);
	      strcpy(devmatch[i].sernum,devlist[i].sernum);
	      devmatch[i].devtype = devlist[i].devtype;
	   }
	   /* Copy the dstatus, even if 0 */
	   devmatch[i].dstatus = devlist[i].dstatus;
	}
        fflush(fdlog);

        if (frunonce) quit(0);
        sleep(interval);
   }                           /*end while(1) */
}  /* end main */

/* end sgdiskmon.c */
